

import json


def loadJson(fileName):
    with open(fileName) as f:
        return json.loads(f.read())


def saveJson(data, fileName):
    with open(fileName, 'w') as f:
        json.dump(data, f, sort_keys=True, indent=4)






import utils



# ---------- Register Model Classes

modelElements = {}

def register(c):
    modelElements[c.__name__] = c
    return c





# ---------- Model Element

class ModelElement:
    def __init__(self, data):
        self._obj = self.__class__.__name__
        if data:
            self.loadData(data)
        else:
            self.loadRandom()

    def loadData(self, data):
        for prop in data.keys():
            setattr(self, prop, makeObjStruct(data[prop]))

    def loadRandom(self):
        pass

    def getData(self):
        return makeDataStruct(self.__dict__)

    def save(self, filePath):
        utils.saveJson(self.getData(), filePath)




def makeObject(data):
    objName = data['_obj']
    objClass = modelElements[objName]
    return objClass(data)


def makeObjStruct(d):
    if type(d) == dict:
        if '_obj' in d.keys():
            return makeObject(d)
        else:
            return {key:makeObjStruct(d[key]) for key in d}
    elif type(d) == list:
        return [makeObjStruct(elem) for elem in d]
    else:
        return d


def makeDataStruct(d):
    if isinstance(d, ModelElement):
        return d.getData()
    elif type(d) == dict:
        return {key:makeDataStruct(d[key]) for key in d}
    elif type(d) == list:
        return [makeDataStruct(elem) for elem in d]
    else:
        return d







import sys
import yaml
import pathlib


from PyQt5 import QtCore
from PyQt5 import QtGui
from PyQt5 import QtWidgets


import rendering
import utils



class SelectElem(QtWidgets.QGroupBox):
    def __init__(self, parent):
        super().__init__("Selection")
        self.parent = parent

        self.elem = None

        self.initUi()

    def initUi(self):
        font = QtGui.QFont("Monospace", 10)

        self.label = QtWidgets.QLabel('None')
        self.label.setFont(font)

        vbox = QtWidgets.QVBoxLayout()
        vbox.addWidget(self.label)

        self.setLayout(vbox)

    def setText(self):
        if self.elem:
            data = self.elem.getData()
            txt = yaml.dump(data, sort_keys=True)
            txt = txt[:-1]
            self.label.setText(txt)
        else:
            self.label.setText('None')

    def setElem(self, elem):
        self.elem = elem

    def update(self):
        self.setText()


class SpeedCtrl(QtWidgets.QGroupBox):
    def __init__(self, parent):
        super().__init__("Speed")
        self.parent = parent

        self.initTimer()
        self.initUi()

    def initTimer(self):
        self.timer = QtCore.QTimer(self)
        self.timer.setInterval(1000)
        self.timer.timeout.connect(self.parent.update)
        self.timer.start()

    def initUi(self):
        slider = QtWidgets.QSlider(QtCore.Qt.Horizontal)
        slider.valueChanged.connect(self.setSpeed)

        self.label = QtWidgets.QLabel('')

        vbox = QtWidgets.QVBoxLayout()
        vbox.addWidget(slider)
        vbox.addWidget(self.label)

        self.setLayout(vbox)

        slider.setValue(10)

    def setSpeed(self, val):
        speed = int(((val)/8)**2)
        self.label.setText(f'{speed}/s')

        if speed == 0:
            self.timer.stop()
        else:
            intervalMs = int(1000/speed)
            self.timer.setInterval(intervalMs)
            self.timer.start()


class ToolBar(QtWidgets.QToolBar):
    def __init__(self, parent):
        super().__init__()

        self.parent = parent

        self.act = {}

        self.addAct('Load', 'load.png', self.load)
        self.addAct('Save', 'save.png', self.parent.save)
        self.addSeparator()
        self.addAct('View', 'view.png', self.parent.resetView)
        self.addSeparator()
        self.addAct('Reset', 'reset.png', self.parent.reset)
        self.addAct('Next', 'next.png', self.parent.update_)
        self.addAct('Play', 'play.png', self.play)
        self.addSeparator()
        self.addAct('Close', 'close.png', self.parent.closeEvent)

    def getIconPath(self, icon):
        return pathlib.Path(__file__).parent / f"icons/{icon}"

    def setIcon(self, act, icon):
        icon = QtGui.QIcon(str(self.getIconPath(icon)))
        self.act[act].setIcon(icon)

    def addAct(self, name, icon, conn):
        icon = QtGui.QIcon(str(self.getIconPath(icon)))
        act = QtWidgets.QAction(icon, name, self)
        act.triggered.connect(conn)
        self.act[name] = act
        self.addAction(act)

    def load(self):
        self.parent.load()
        if self.parent.model:
            self.act['Save'].setEnabled(True)

    def play(self):
        if self.parent.state == 'pause':
            self.parent.state = 'play'
            self.setIcon('Play', 'pause.png')
        else:
            self.parent.state = 'pause'
            self.setIcon('Play', 'play.png')



class MainWindow(QtWidgets.QMainWindow):
    def __init__(self, modelClass, filePath=None):
        super().__init__()

        self.modelClass = modelClass
        self.filePath = filePath
        self.model = None
        self.state = 'pause'

        self.resize(800, 500)
        self.setWindowTitle('Ecofaber')

        self.glWidget = rendering.GLWidget(self)
        self.initGUI()
        self.show()

        if self.filePath:
            self.load(self.filePath)

    def initCtrl(self):
        self.speedCtrl = SpeedCtrl(self)
        self.selectElem = SelectElem(self)

        ctrlLayout = QtWidgets.QVBoxLayout()
        ctrlLayout.addWidget(self.speedCtrl)
        ctrlLayout.addWidget(self.selectElem)
        ctrlLayout.addStretch()

        self.ctrlWidget = QtWidgets.QWidget()
        self.ctrlWidget.setLayout(ctrlLayout)

    def initGUI(self):
        self.initCtrl()

        self.toolBar = ToolBar(self)
        self.addToolBar(self.toolBar)

        guiLayout = QtWidgets.QHBoxLayout()
        guiLayout.addWidget(self.glWidget)
        guiLayout.addWidget(self.ctrlWidget)

        guiLayout.setStretch(0, 4)
        guiLayout.setStretch(1, 1)

        centralWidget = QtWidgets.QWidget()
        centralWidget.setLayout(guiLayout)
        self.setCentralWidget(centralWidget)

    def resetView(self):
        self.glWidget.resetView()
        self.updateView()

    def reset(self):
        if self.filePath:
            data = utils.loadJson(self.filePath)
            self.model = self.modelClass(data)
            self.selectElem.setElem(None)
            self.updateView()

    def load(self, filePath=None):
        if filePath == None:
            dFun = QtWidgets.QFileDialog.getOpenFileName
            filePath, _ = dFun(self, 'Open project', '.', 'Json (*.json)')

        if filePath:
            self.filePath = filePath
            self.reset()
            self.resetView()

    def save(self):
        dFun = QtWidgets.QFileDialog.getSaveFileName
        filePath, _ = dFun(self, 'Save project', '.', 'Json (*.json)')

        if filePath:
            if filePath[-5:] != '.json':
                filePath += '.json'
            self.model.save(filePath)

    def updateView(self):
        self.glWidget.updateGL()
        self.selectElem.update()

    def update_(self):
        if self.model:
            self.model.update()
            self.updateView()

    def update(self):
        if self.state == 'play':
            self.update_()

    def closeEvent(self, event):
        sys.exit()

    def setSelectElem(self, elem):
        self.selectElem.setElem(elem)
        self.updateView()




class App(QtWidgets.QApplication):
    def __init__(self, modelClass, modelCfgPath=None):
        super().__init__([])

        self.win = MainWindow(modelClass, modelCfgPath)

        sys.exit(self.exec_())







import sys
import numpy as np


from PyQt5 import QtCore
from PyQt5 import QtGui
from PyQt5 import QtWidgets
from PyQt5 import QtOpenGL

import OpenGL.GL as gl
from OpenGL import GLU


from stl import mesh



class Shape3D:
    def __init__(self, filePath):
        self.mesh = mesh.Mesh.from_file(filePath)

        v, c, i = self.mesh.get_mass_properties()

        self.volume = v
        self.cog = c
        self.selRadius = self.getSelRadius(v)

    def getSelRadius(self, volume):
        return np.power(volume, 1/3)*0.6

    def getPoints(self):
        return self.mesh.points



class Camera:
    def __init__(self):
        self.action = None

        self.eye = np.array([100.,0.,0.])
        self.ctr = np.array([0.,0.,0.])
        self.up = np.array([0.,0.,1.])

    def getOrth(self):
        n = self.eye - self.ctr
        n = n / np.linalg.norm(n)
        j = self.up - np.dot(self.up, n)*n
        j = j / np.linalg.norm(j)
        i = np.cross(n, j)

        return n, i, j

    def getEulerAngles(self, v):
        v = v / np.linalg.norm(v)
        a = np.arctan2(v[1],v[0])
        b = np.arcsin(v[2])
        return a, b

    def getNormal(self, a, b):
        if b > 1.55:
            b = 1.55
        if b < -1.55:
            b = -1.55
        return np.array([
            np.cos(a)*np.cos(b),
            np.sin(a)*np.cos(b),
            np.sin(b)
        ])

    def setAction(self, action):
        self.action = action

    def setScroll(self, scroll):
        n, _ , _ = self.getOrth()
        self.eye -= + n*scroll

    def setMove(self, move):
        n, i, j = self.getOrth()

        if self.action == 'move':
            move = move / 10
            self.eye += (move[0]*i - move[1]*j)
            self.ctr += (move[0]*i - move[1]*j)

        if self.action == 'rotate':
            move = move / 100

            a, b = self.getEulerAngles(n)
            n = self.getNormal(a - move[0], b - move[1])

            dist = np.linalg.norm(self.eye - self.ctr)
            self.eye = self.ctr + n*dist

    def getState(self):
        return list(self.eye) + list(self.ctr) + list(self.up)

    def getRay(self, coord, angle):
        n, _ , _ = self.getOrth()
        a, b = self.getEulerAngles(-n)

        ar = np.arctan(coord[0]*np.tan(np.radians(angle[1]/2)))
        br = np.arctan(coord[1]*np.tan(np.radians(angle[0]/2)))

        a -= ar/np.cos(b)
        b += br

        ro = self.eye
        rn = self.getNormal(a, b)

        return ro, rn



class RayCasting:
    def __init__(self):
        pass

    def pick(self, ray, objects, shapes):
        ro, rn = ray
        matches = {}

        for obj in objects:
            if 'sel' not in obj.keys() or obj['sel'] == False:
                continue

            selSize = shapes[obj['name']].selRadius
            pos = np.array(obj['pos'])

            b = np.dot(rn, ro - pos)
            c = np.dot(ro - pos, ro - pos) - np.square(selSize)
            det = np.square(b) - c

            if det  <= 0:
                continue

            s = np.sqrt(det)
            t1, t2 = -b + s, -b -s

            for t in (t1, t2):
                if t > 0:
                    matches[t] = obj

        if len(matches.keys()):
            firstT = sorted(matches.keys())[0]
            return matches[firstT]
        else:
            return None



class GLWidget(QtOpenGL.QGLWidget):
    def __init__(self, parent=None):
        self.parent = parent
        QtOpenGL.QGLWidget.__init__(self, parent)

        self.shapes = {}
        self.camera = Camera()
        self.ray = None

    def initializeGL(self):
        self.qglClearColor(QtGui.QColor(155, 155, 155))
        gl.glEnable(gl.GL_DEPTH_TEST)

    def resizeGL(self, width, height):
        self.size = [width, height]
        gl.glViewport(0, 0, width, height)
        gl.glMatrixMode(gl.GL_PROJECTION)
        gl.glLoadIdentity()
        aspect = width / float(height)
        GLU.gluPerspective(25.0, aspect, 1.0, 1000.0)
        gl.glMatrixMode(gl.GL_MODELVIEW)

    def addShape(self, name):
        if name not in self.shapes.keys():
            self.shapes[name] = Shape3D(f'shapes/{name}.stl')
        return self.shapes[name]

    def paintObject(self, obj, col, pos):
        gl.glPushMatrix()

        gl.glTranslate(*pos)
        gl.glTranslate(*[-i for i in obj.cog])

        gl.glColor3ub(*col)
        gl.glBegin(gl.GL_TRIANGLES)
        for p in obj.getPoints():
            gl.glVertex3f(*p[0:3])
            gl.glVertex3f(*p[3:6])
            gl.glVertex3f(*p[6:9])
        gl.glEnd()

        gl.glPopMatrix()

    def paintGL(self):
        gl.glClear(gl.GL_COLOR_BUFFER_BIT | gl.GL_DEPTH_BUFFER_BIT)
        gl.glLoadIdentity()

        GLU.gluLookAt(*self.camera.getState())

        model = self.parent.model
        if model:
            for obj in model.getView():
                shape = self.addShape(obj['name'])
                self.paintObject(shape, obj['color'], obj['pos'])

        '''
        if self.ray:
            ro, rn = self.ray
            pos = list(ro + 100*rn)
            shape = self.addShape('b_tetra_1')
            self.paintObject(shape, [255,255,255], pos)
        '''


    def resetView(self):
        self.camera = Camera()
        self.updateGL()

    def selectObject(self):
        coord = [
            (2. * self.mousePos[0]) / self.size[0] - 1.,
            (2. * self.mousePos[1]) / self.size[1] + 1.
        ]

        angle = [
            25.,
            25.*self.size[0]/self.size[1]
        ]

        self.ray = self.camera.getRay(coord, angle)
        objects = self.parent.model.getView()

        raycasting = RayCasting()
        res = raycasting.pick(self.ray, objects, self.shapes)
        if res:
            return self.parent.model.getObj(res['pos'])
        else:
            return None

    def mousePressEvent(self, event):
        self.mousePos = np.array([event.x(), -event.y()])

        obj = self.selectObject()
        if obj:
            self.parent.setSelectElem(obj)

        btnNb = event.button()

        actions = {
            1: 'move',  # left click
            2: 'rotate' # right click
        }

        if btnNb in actions.keys():
            action = actions[btnNb]
            self.camera.setAction(action)
            self.updateGL()

    def mouseReleaseEvent(self, event):
        self.camera.setAction(None)

    def mouseMoveEvent(self, event):
        pos = np.array([event.x(), -event.y()])
        move = pos - self.mousePos
        self.mousePos = pos

        self.camera.setMove(move)
        self.updateGL()

    def wheelEvent(self, event):
        self.camera.setScroll(event.angleDelta().y()/120)
        self.updateGL()

